MODULE Files;

IMPORT
  SYSTEM, Trace, Kernel32;

CONST
  pathChar*    = "/";

TYPE
  String       = ARRAY 255 OF CHAR;
  Bytes4       = ARRAY 4 OF SYSTEM.BYTE;
  Bytes8       = ARRAY 8 OF SYSTEM.BYTE;
  FileName*    = String;

  File* = OBJECT
  VAR
    handle : LONGINT;
    name*  : String;
    path*  : String;
    len*   : LONGINT;
    pos*   : LONGINT;
    eof*   : BOOLEAN;

    PROCEDURE & Init*;
    BEGIN
      pos  := 0;
      len  := 0;
      eof  := FALSE;
    END Init;

    PROCEDURE Open*(filename: String): BOOLEAN;
    VAR
      res: LONGINT;
    BEGIN
      name := filename;
      path := filename;
      IF Exists(path) THEN
        handle := Kernel32.CreateFile(
	            path,
		    Kernel32.FILE_GENERIC_READ + Kernel32.FILE_GENERIC_WRITE,
                    Kernel32.FILE_SHARE_READ+Kernel32.FILE_SHARE_WRITE+Kernel32.FILE_SHARE_DELETE, NIL,
		    Kernel32.OPEN_EXISTING, Kernel32.FILE_ATTRIBUTE_NORMAL, 0);
        SetPos(0);
        len := Kernel32.GetFileSize(handle,res);
        RETURN TRUE;
      ELSE
        Trace.Error("'"); Trace.Error(name); Trace.Error("' ");
        Trace.Error('File not existed!');
        eof  := TRUE;
        RETURN FALSE;
      END
    END Open;

    PROCEDURE Create*(filename: String): BOOLEAN;
    VAR
      res: LONGINT;
    BEGIN
      handle := Kernel32.CreateFile(
                  path,
                  Kernel32.FILE_GENERIC_READ + Kernel32.FILE_GENERIC_WRITE,
                  {}, NIL,
                  Kernel32.CREATE_NEW,Kernel32.FILE_ATTRIBUTE_NORMAL, 0);
      pos := 0;
      len := 0;
      IF handle = Kernel32.INVALID_HANDLE_VALUE THEN RETURN FALSE; END;
      RETURN TRUE;
    END Create;

    PROCEDURE Close*;
    BEGIN
      IF handle # Kernel32.INVALID_HANDLE_VALUE THEN
        Kernel32.CloseHandle(handle);
      END;
    END Close;

    PROCEDURE SetPos*(position : LONGINT);
    VAR
      res : LONGINT;
    BEGIN
      Kernel32.SetFilePointer(handle,position,res,Kernel32.FILE_BEGIN);
      pos := position;
    END SetPos;

    PROCEDURE Length*(): LONGINT;
    BEGIN
      RETURN len;
    END Length;

    PROCEDURE ReadByte*(VAR x: SYSTEM.BYTE);
    VAR
      res,count : LONGINT;
      b: ARRAY 1 OF SYSTEM.BYTE;
    BEGIN
      IF pos < len THEN
        Kernel32.ReadFile(handle,b,1,count,NIL);
        INC(pos);
        x := b[0];
      ELSE
        x := 0;
        eof := TRUE
      END
    END ReadByte;

    PROCEDURE ReadBytes*(VAR x: ARRAY OF SYSTEM.BYTE; len: LONGINT);
    VAR
      b: SYSTEM.BYTE;
      i: LONGINT;
    BEGIN
      IF SELF.pos+len<SELF.len THEN
        Kernel32.ReadFile(handle,x,len,i,NIL);
        INC(pos,len);
      ELSIF SELF.len-SELF.pos>0 THEN
        Kernel32.ReadFile(handle,x,SELF.len-SELF.pos,i,NIL);
        INC(pos,SELF.len-SELF.pos);
      ELSE
        Trace.Error('Length to long');
      END;
    END ReadBytes;

    PROCEDURE Read*(VAR x: SYSTEM.BYTE);
    BEGIN
      ReadByte(x);
    END Read;

    PROCEDURE ReadInt*(VAR x: INTEGER);
    VAR
      x0, x1: SHORTINT;
    BEGIN
      Read(x0);
      Read(x1);
      x := LONG(x1) * 100H + LONG(x0) MOD 100H;
    END ReadInt;

    PROCEDURE ReadLInt*(VAR x: LONGINT);
    BEGIN
      ReadBytes(SYSTEM.VAL(Bytes4, x), 4)
    END ReadLInt;

    PROCEDURE ReadSet*(VAR x: SET);
    BEGIN
      ReadBytes(SYSTEM.VAL(Bytes4, x), 4)
    END ReadSet;

    PROCEDURE ReadBool*(VAR x: BOOLEAN);
    VAR
      s: SHORTINT;
    BEGIN
      Read(s);
      x := s # 0
    END ReadBool;

    PROCEDURE ReadReal*(VAR x: REAL);
    BEGIN
      ReadBytes(SYSTEM.VAL(Bytes4, x), 4)
    END ReadReal;

    PROCEDURE ReadLReal*(VAR x: LONGREAL);
    BEGIN
      ReadBytes(SYSTEM.VAL(Bytes8, x), 8)
    END ReadLReal;

    PROCEDURE ReadString*(VAR x: ARRAY OF CHAR);
    VAR
      i: LONGINT;
      ch: CHAR;
    BEGIN
      i := 0;
      REPEAT
        Read(ch);
        x[i] := ch;
        INC(i)
      UNTIL ch = 0X;
    END ReadString;

    PROCEDURE ReadNum*(VAR x: LONGINT);
    VAR
      ch: CHAR;
      n, y: LONGINT;
    BEGIN
      n := 0;
      y := 0;
      Read(ch);
      WHILE ch >= 80X DO
        INC(y, SYSTEM.LSH(LONG(ch) - 128, n));
        INC(n, 7);
        Read(ch)
      END;
      x := ASH(SYSTEM.LSH(LONG(ch), 25), n-25) + y;
    END ReadNum;

  END File;

PROCEDURE Exists*(path: String): BOOLEAN;
VAR
  res: SET;
BEGIN
  res := Kernel32.GetFileAttributes(path);
  RETURN res # Kernel32.INVALID_FILE_ATTRIBUTES;
END Exists;

END Files.




