MODULE Classes;

IMPORT
  SYSTEM, W := Windows;

CONST
  noClass*     = 0;
  sysClass*    = 1;
  userClass*   = 2;

TYPE
  Char         = W.Char;
  Boolean      = W.Boolean;
  Shortint     = W.Shortint;
  Integer      = W.Integer;
  Longint      = W.Longint;
  Real         = W.Real;
  Longreal     = W.Longreal;
  Pointer      = W.Pointer;
  Set	       = W.Set;
  String       = W.String;
  Name         = W.Name;

  ObjectArray        = POINTER TO ARRAY OF Pointer;

  List* = POINTER TO RECORD
    list*: ObjectArray;
    count-: Longint;

    PROCEDURE & InitList*(initialSize: Longint) ;
    BEGIN
      IF initialSize <= 0 THEN initialSize := 8 END;
      NEW(list,initialSize);
      count := 0;
    END InitList;

    PROCEDURE Length*( ): Longint;
    BEGIN
      RETURN count
    END Length;

    PROCEDURE Grow;
    VAR
      old: ObjectArray;
      i: Longint;
    BEGIN
      old := list;
      NEW(list, (LEN(list) * 3+1) DIV 2);
      FOR i := 0 TO count - 1 DO list[i] := old[i] END
    END Grow;

    PROCEDURE Add*(x: Pointer);
    BEGIN
      IF count = LEN(list) THEN Grow END;
      list[count] := x;
      INC(count)
    END Add;

    PROCEDURE Remove*(x: Pointer);
    VAR
      i: Longint;
    BEGIN
      i := 0;
      WHILE (i < count) & (list[i] # x) DO INC( i ) END;
      IF i < count THEN
	WHILE (i < count - 1) DO
          list[i] := list[i + 1];
          INC( i )
        END;
	DEC(count);
        list[count] := NIL;
      END
    END Remove;

    PROCEDURE Insert*(i: Longint; x: Pointer);
    VAR
      j: Longint;
    BEGIN
      IF count = LEN(list) THEN Grow END;
      INC(count);
      j := count - 2;
      WHILE (j >= i) DO
        list[j+1] := list[j];
        DEC(j);
      END;
      list[i] := x;
    END Insert;

    PROCEDURE IndexOf*(x: Pointer): Longint;
    VAR
      i: Longint;
    BEGIN
      i := 0;
      WHILE i < count DO
        IF list[i] = x THEN RETURN i END;
	INC(i);
      END;
      RETURN -1
    END IndexOf;

    PROCEDURE Clear*;
    VAR
      i: Longint;
    BEGIN
      FOR i := 0 TO count - 1 DO list[i] := NIL END;
      count := 0;
    END Clear;

  END;

  StringList*       = POINTER TO RECORD
    left   : StringList;
    right  : StringList;
    name   : Name;
    sys    : Boolean;
    count* : Longint;

    PROCEDURE & Init*;
    BEGIN
      left    := NIL;
      right   := NIL;
      name    := "";
      count   := 0;
    END Init;

    PROCEDURE Add*(name: String; sys: Boolean);
    VAR
      ob0, ob1: StringList;
      left: Boolean;
      res: Name;
      str : ARRAY 255 OF Char;
    BEGIN

      ob0 := SELF;
      ob1 := ob0.right;
      left := FALSE;

      UpperCase(name, res);

      LOOP
        IF ob1 # NIL THEN
          IF res < ob1.name THEN
	    ob0 := ob1;
	    ob1 := ob0.left;
	    left := TRUE
          ELSIF res > ob1.name THEN
	    ob0 := ob1;
	    ob1 := ob0.right;
	    left := FALSE
          ELSE
            EXIT
          END;
        ELSE (*insert*)
          NEW(ob1);
          IF left THEN
	    ob0.left := ob1
          ELSE
	    ob0.right := ob1
          END;
          ob1.left:=NIL;
          ob1.right:=NIL;
          ob1.name := res;
          ob1.sys := sys;
          INC(count);
          EXIT;
        END;
      END;

    END Add;

    PROCEDURE Find*(name: String): Shortint;
    VAR
      obj: StringList;
      res: Name;
    BEGIN
      UpperCase(name, res);
      obj := SELF;
      LOOP
        IF obj # NIL THEN
          IF res < obj.name THEN
	    obj := obj.left;
          ELSIF res > obj.name THEN
	    obj := obj.right;
          ELSE (*ok*)
            EXIT
          END;
        ELSE (*no found*)
          RETURN noClass;
        END;
      END;
      IF obj.sys THEN
        RETURN sysClass;
      ELSE
        RETURN userClass;
      END;
    END Find;

    PROCEDURE UnregisterClass*;

      PROCEDURE Show(VAR obj: StringList);
      BEGIN
        IF obj = NIL THEN RETURN END;
        IF obj.left  # NIL THEN Show(obj.left); END;
        IF obj.right # NIL THEN Show(obj.right); END;
        IF ~obj.sys THEN W.UnregisterClass(obj.name,W.hInstance); END;
      END Show;

    BEGIN
      Show(left);
      Show(right);
    END UnregisterClass;

  END;

PROCEDURE Length* (s: String): Longint;
VAR
  i: Longint;
BEGIN
  i := 0;
  WHILE (i < LEN(s)) & (s[i] # 0X) DO i := i + 1 END;
  RETURN i;
END Length;

PROCEDURE Insert* (source: String; pos: Longint; VAR dest: String);
VAR
  srclen, dstlen, i, k : Longint;
BEGIN
  dstlen := Length(dest);
  IF (pos >= 0) & (pos <= dstlen) THEN
    IF pos = dstlen THEN
      i := 0;
      WHILE (pos + i < LEN(dest) - 1) & (source[i] # 0X) DO
        dest[pos + i] := source[i];
        i := i + 1;
      END;
      dest[pos + i] := 0X;
    ELSE
      srclen := Length(source);
      IF srclen > 0 THEN
        k := dstlen + srclen - 1;
        IF k > LEN(dest) - 2 THEN k := LEN(dest) - 2 END;
        dest[k + 1] := 0X; i := k;
        WHILE i - srclen >= pos DO
          dest[i] := dest[i - srclen];
          i := i - 1;
        END;
        i := 0;
        WHILE (i < srclen) & (pos + i <= k) DO
          dest[pos + i] := source[i];
          i := i + 1;
        END;
      END;
    END;
  END;
END Insert;

PROCEDURE Append* (extra: String; VAR dest: String);
BEGIN
  Insert (extra, Length(dest), dest);
END Append;

PROCEDURE Lowercase*(name: String; VAR res: String);
VAR
  ch: Char;
  i : Longint;
BEGIN
  i := 0;
  REPEAT
    ch := name[i];
    IF (ch >= 'A') & (ch <= 'Z') THEN
      ch := CHR(ORD(ch)-ORD('A')+ORD('a'));
    END;
    res[i] := ch;
    INC(i);
  UNTIL ch = 0X;
END Lowercase;

PROCEDURE UpperCase*(name: String; VAR res: String);
VAR
  i: Longint;
BEGIN
  i:=0;
  WHILE  (name[i] # 0X) DO
    CASE name[i] OF
      "a".."z": res[i] := CAP(name[i]);
    ELSE
      res[i] := name[i];
    END;
    INC(i);
    res[i] := 0X;
  END;
END UpperCase;

PROCEDURE IntToStr*(val: Longint): String;
VAR
  i, j  : Longint;
  str   : ARRAY 16 OF Char;
  digits: ARRAY 16 OF Longint;
BEGIN
    IF val = MIN(Integer) THEN
       COPY("-2147483648", str);
       RETURN str
    END;
    IF val < 0 THEN
       val := -val;
       str[0] := "-";
       j := 1
    ELSE
       j := 0
    END;
    i := 0;
    REPEAT
       digits[i] := val MOD 10;
       INC(i);
       val := val DIV 10
    UNTIL val = 0;
    DEC(i);
    WHILE i >= 0 DO
       str[j] := CHR(digits[i]+ORD("0"));
       INC(j);
       DEC(i)
    END;
    str[j] := 0X;
    RETURN str;
END IntToStr;

VAR
  ClassList* : StringList;

BEGIN
  NEW(ClassList);

  ClassList.Add("Button"	    ,TRUE);
  ClassList.Add("ComboBox"	    ,TRUE);
  ClassList.Add("Edit"		    ,TRUE);
  ClassList.Add("ListBox"	    ,TRUE);
  ClassList.Add("MDIClient"	    ,TRUE);
  ClassList.Add("ScrollBar"	    ,TRUE);
  ClassList.Add("Static"	    ,TRUE);

  ClassList.Add("ToolbarWindow32"   ,TRUE);
  ClassList.Add("ReBarWindow32"	    ,TRUE);
  ClassList.Add("ComboBoxEx32"	    ,TRUE);
  ClassList.Add("tooltips_class32"  ,TRUE);

  ClassList.Add("SysListView32"	    ,TRUE);
  ClassList.Add("SysTreeView32"	    ,TRUE);
  ClassList.Add("SysHeader32"	    ,TRUE);
  ClassList.Add("SysTabControl32"   ,TRUE);
  ClassList.Add("SysAnimate32"	    ,TRUE);
  ClassList.Add("SysMonthCal32"	    ,TRUE);
  ClassList.Add("SysDateTimePick32" ,TRUE);
  ClassList.Add("SysIPAddress32"    ,TRUE);

  ClassList.Add("msctls_statusbar32",TRUE);
  ClassList.Add("msctls_trackbar32" ,TRUE);
  ClassList.Add("msctls_updown32"   ,TRUE);
  ClassList.Add("msctls_progress32" ,TRUE);
  ClassList.Add("msctls_hotkey32"   ,TRUE);
END Classes.
