MODULE Controls;

IMPORT
  SYSTEM, Trace, W := Windows, Cl := Classes, G := Graphics;

CONST
  alNone*       = 0;
  alLeft*       = 1;
  alTop*        = 2;
  alRight*      = 3;
  alBottom*     = 4;
  alClient*     = 5;

  poCustom*        = 0;
  poLeftTop*       = 1;
  poLeftCenter*    = 2;
  poLeftBottom*    = 3;
  poCenterTop*     = 4;
  poCenterCenter*  = 5;
  poCenterBottom*  = 6;
  poRightTop*      = 7;
  poRightCenter*   = 8;
  poRightBottom*   = 9;

  mbLeft*       = 1;
  mbRight*      = 2;
  mbMiddle*     = 3;

VAR
  terminated*  : Boolean;
  mainform*    : Control;
  mainhint*    : Hint;
  screen*      : List;

TYPE
  Char         = W.Char;
  Boolean      = W.Boolean;
  Shortint     = W.Shortint;
  Integer      = W.Integer;
  Longint      = W.Longint;
  Real         = W.Real;
  Longreal     = W.Longreal;
  Pointer      = W.Pointer;
  Set	       = W.Set;
  String       = W.String;
  Name         = W.Name;

  NotifyEvent* = PROCEDURE {DELEGATE} (Sender: Object);

  List = POINTER TO RECORD(Cl.List)
    width* : Longint;
    height*: Longint;
  END;

  Message* = RECORD
    msg*    : Longint;
    wparam* : Longint;
    lparam* : Longint;
    result* : Longint;
    button* : Shortint;
    dc*     : Longint;
  END;

  State*    = POINTER TO RECORD
    left*      : Longint;
    top*       : Longint;
    width*     : Longint;
    height*    : Longint;
    color*     : Longint;
    visible*   : Boolean;
    enabled*   : Boolean;
    align*     : Longint;
    position*  : Shortint;
  END;

  Object* = POINTER TO RECORD
    classname*: Name;
  END;

  Persistent* = POINTER TO RECORD (Object)
    name*           : Name;
    tag*            : Longint;
    PROCEDURE Destroy*;
    BEGIN
    END Destroy;
  END;

  Control* = POINTER TO RECORD (Persistent)

    owner*          : Control;
    handle*         : Longint;
    caption*        : Name;
    controls*       : Cl.List;

    left*           : Longint;
    top*            : Longint;
    width*          : Longint;
    height*         : Longint;
    color*          : Longint;
    visible*        : Boolean;
    enabled*        : Boolean;

    style*          : Set;
    exstyle*        : Set;
    wndstyle*       : Set;
    wndparent*      : Longint;
    wndparam*       : Longint;
    idcontrol*      : Longint;

    state*          : State;
    statedata*      : Set;

    align*          : Longint;
    border*         : Longint;
    position*       : Shortint;
    tabstop*        : Boolean;

    hint*           : Name;
    showhint*       : Boolean;
    iconname*       : Name;

    font*           : G.Font;

    brush           : Longint;
    mouseinside     : Boolean;
    mousehover      : Boolean;
    mousedown*      : Boolean;
    forhover        : W.Point;

    OnClick*        : NotifyEvent;

    PROCEDURE & Create*(AOwner: Control);
    VAR
      rect: W.Rect;
    BEGIN
      classname      := "Control";
      caption        := "o.b.e.r.o.n";

      left           := 0;
      top            := 0;
      width          := 25;
      height         := 25;

      visible        := TRUE;
      enabled        := TRUE;
      handle         := 0;
      color          := G.clBtnFace;

      wndparent      := 0;
      idcontrol      := 0;
      wndparam       := 0;
      showhint       := FALSE;
      iconname       := "";

      NEW(font,"Tahoma");

      IF AOwner # NIL THEN
        AOwner.Insert(SELF);
        showhint := AOwner.showhint;
        color    := AOwner.color;
      ELSE
        IF screen = NIL THEN
          NEW(screen,0);
          W.SystemParametersInfo(W.SPI_GETWORKAREA,0,SYSTEM.ADR(rect),0);
          screen.width  := rect.right - rect.left;
          screen.height := rect.bottom - rect.top;
        END;
        screen.Add(SELF);
        IF mainform = NIL THEN
          mainform := SELF;
          IF mainhint = NIL THEN NEW(mainhint,NIL) END;
        END;
      END;

      align          := alNone;
      position       := poCustom;
      border         := 0;
      brush          := 0;
      mouseinside    := FALSE;
      mousehover     := FALSE;
      mousedown      := FALSE;

      tabstop        := FALSE;
      hint           := "";
    END Create;

    PROCEDURE Destroy*;
    VAR
      i: Longint;
    BEGIN
      IF controls # NIL THEN
        FOR i := 0 TO controls.count - 1 DO
          IF controls.list[i](Control) # NIL THEN
            controls.list[i](Control).Destroy;
          END;
        END;
        controls.Clear;
        controls := NIL;
      END;
      Destroy^;
    END Destroy;

    PROCEDURE Insert(AControl: Control);
    BEGIN
      IF controls = NIL THEN NEW(controls,0) END;
      controls.Add(AControl);
      AControl.owner := SELF;
    END Insert;

    PROCEDURE Remove(AControl: Control);
    BEGIN
      AControl.owner := NIL;
      IF controls # NIL THEN
        controls.Remove(AControl);
        IF controls.count = 0 THEN controls := NIL END;
      END;
    END Remove;

    PROCEDURE CreateParams*;
    BEGIN
      style     := W.WS_CLIPCHILDREN + W.WS_CHILD + W.WS_CLIPSIBLINGS (*+ W.WS_BORDER*);
      exstyle   := W.WS_EX_CONTROLPARENT;
      wndstyle  := {};
      IF owner # NIL THEN wndparent := owner.handle END;
    END CreateParams;

    PROCEDURE CreateHandle;
    BEGIN
      handle := W.CreateWindowEx(
        exstyle, classname,
        caption, style,
        left, top, width, height,
        wndparent, idcontrol, W.hInstance,
        wndparam
      );
      IF handle # 0 THEN
        W.SetProp(handle,SYSTEM.ADR("Control"),SYSTEM.VAL(Longint,SELF));
        IF state = NIL THEN NEW(state) END;
      ELSE
      END;
    END CreateHandle;

    PROCEDURE PrecedingWindow(control: Control): Longint;
    VAR
      i     : Longint;
      Result: Longint;
    BEGIN
      FOR i := controls.IndexOf(control) + 1 TO controls.count - 1 DO
        Result := controls.list[i](Control).handle;
        IF Result # 0 THEN RETURN Result END;
      END;
      RETURN W.HWND_TOP;
    END PrecedingWindow;

    PROCEDURE UpdateBounds;
    VAR
      R : W.Rect;
    BEGIN
      IF owner = NIL THEN
        W.GetWindowRect(handle,R);
        W.AdjustWindowRectEx(R,{},0,exstyle);
        left := R.left;
        top  := R.top;
      ELSE
        W.GetClientRect(handle,R);
        W.AdjustWindowRectEx(R,style,0,exstyle);
      END;
      width  := R.right-R.left;
      height := R.bottom-R.top;
    END UpdateBounds;

    PROCEDURE SetPosition;
    BEGIN
      CASE position OF
      |  poLeftTop       : left :=0; top :=0;
      |  poLeftCenter    : left :=0; top := (screen.height - height) DIV 2;
      |  poLeftBottom    : left :=0; top := screen.height - height;
      |  poCenterTop     : left := (screen.width - width) DIV 2; top := 0;
      |  poCenterCenter  : left := (screen.width - width) DIV 2; top := (screen.height - height) DIV 2;
      |  poCenterBottom  : left := (screen.width - width) DIV 2; top := screen.height - height;
      |  poRightTop      : left := screen.width - width; top := 0;
      |  poRightCenter   : left := screen.width - width; top := (screen.height - height) DIV 2;
      |  poRightBottom   : left := screen.width - width; top := screen.height - height;
      ELSE
      END;
    END SetPosition;

    PROCEDURE CreateWindow(): Boolean;
    CONST
      maxcontrol = 500;
    VAR
      class   : W.WndClassEx;
      typ     : Shortint;
      tmpname : Name;
      parent  : Control;
      res     : Boolean;

      PROCEDURE CreateName;
      VAR
        i: Longint;
      BEGIN
        i := 1;
        tmpname := classname;
        Cl.Append(Cl.IntToStr(i),tmpname);
        typ := Cl.ClassList.Find(tmpname);
        WHILE (typ # Cl.noClass) & (i<maxcontrol) DO
          INC(i);
          tmpname := classname;
          Cl.Append(Cl.IntToStr(i),tmpname);
          typ := Cl.ClassList.Find(tmpname);
        END;
        name := tmpname;
      END CreateName;

    BEGIN
      CreateParams;

      IF ~ enabled THEN
        style := style + W.WS_DISABLED;
      END;
      IF tabstop THEN style := style + W.WS_TABSTOP END;

      typ := Cl.ClassList.Find(classname);
      IF typ = Cl.noClass THEN
        CreateName;
        classname := tmpname;
      ELSIF typ = Cl.sysClass THEN
        CreateName;
        Cl.ClassList.Add(tmpname,TRUE);
        typ := Cl.sysClass;
      END;

      IF caption = "o.b.e.r.o.n" THEN caption := name END;

      IF typ = Cl.noClass THEN
        IF brush # 0 THEN W.DeleteObject(brush) END;
        brush               := W.CreateSolidBrush(color);
        class.cbSize        := SIZE(W.WndClassEx);
        class.style         := wndstyle;
        class.lpfnWndProc   := WndProc;
        class.cbClsExtra    := 0;
        class.cbWndExtra    := 0;
        class.hInstance     := W.hInstance;
        IF iconname = "" THEN
          class.hIcon         := W.LoadIcon(0,SYSTEM.VAL(W.LPSTR,1));
        ELSE
          class.hIcon         := W.LoadImage(0, SYSTEM.ADR(iconname), W.IMAGE_ICON, 0, 0, W.LR_LOADFROMFILE);
        END;
        class.hIconSm       := W.LoadIcon(0,SYSTEM.VAL(W.LPSTR,1));
        class.hCursor       := W.LoadCursor(0,SYSTEM.VAL(W.LPSTR,W.IDC_ARROW));
        class.hbrBackground := brush;
        class.lpszMenuName  := W.NULL;
        class.lpszClassName := SYSTEM.ADR(classname);

        W.RegisterClassEx(class);
        Cl.ClassList.Add(classname,FALSE);
      ELSIF typ = Cl.sysClass THEN
        IF owner # NIL THEN
          IF owner.controls # NIL THEN
            idcontrol := owner.controls.IndexOf(SELF);
          END;
        END;
      END;

      IF (wndparent = 0) & (style * W.WS_CHILD # {}) THEN
        W.MessageBox(0,"wndparent = 0 & style = W.WS_CHILD","Error!",W.MB_ICONERROR+W.MB_OK);
        IF owner = NIL THEN
          screen.Remove(SELF);
        ELSE
          owner.Remove(SELF);
        END;
        RETURN FALSE;
      END;

      BeforeCreate;
      CreateHandle;

      IF handle # 0 THEN

        AfterCreate;
        UpdateBounds;

        IF owner # NIL THEN
          W.SetWindowPos(
            handle, owner.PrecedingWindow(SELF),
            0, 0, 0, 0,
            W.SWP_NOMOVE + W.SWP_NOSIZE + W.SWP_NOACTIVATE
          );
        END;

      ELSE
        W.MessageBox(0,"handle = 0","Error!",W.MB_ICONERROR+W.MB_OK);
        IF owner = NIL THEN
          screen.Remove(SELF);
        ELSE
          owner.Remove(SELF);
        END;
        RETURN FALSE;
      END;

      RETURN TRUE;

    END CreateWindow;

    PROCEDURE CreateControl*;
    VAR
      i,param: Longint;
    BEGIN
      IF handle = 0 THEN
        IF CreateWindow() THEN
          IF controls # NIL THEN
            FOR i := 0 TO controls.count - 1 DO
              controls.list[i](Control).CreateControl;
              controls.list[i](Control).top := controls.list[i](Control).top + border;
              controls.list[i](Control).left := controls.list[i](Control).left + border;
            END;
          END;
          IF visible THEN
            param := W.SW_SHOW;
          ELSE
            param := W.SW_HIDE;
          END;
          AlignControls;
          W.ShowWindow(handle,param);
        ELSE
          W.ExitProcess(0);
        END;
      END;
    END CreateControl;

    PROCEDURE AlignControls;
    VAR
      i           : Longint;
      lw,th,rw,bh : Longint;
      rect        : W.Rect;
      leftrect    : W.Rect;
      toprect     : W.Rect;
      rightrect   : W.Rect;
      bottomrect  : W.Rect;
      clientrect  : W.Rect;

      newleft     : Longint;
      newtop      : Longint;
      newwidth    : Longint;
      newheight   : Longint;

      flag        : Boolean;

      PROCEDURE AdjustRects;
      VAR
        i: Longint;
      BEGIN
        leftrect  := rect;
        leftrect.right := leftrect.left;
        toprect   := rect;
        toprect.bottom := toprect.top;
        rightrect := rect;
        rightrect.left := rightrect.right;
        bottomrect:= rect;
        bottomrect.top := bottomrect.bottom;
        clientrect:= rect;
        FOR i := controls.count - 1 TO 0 BY -1 DO
          CASE controls.list[i](Control).align OF
          |  alLeft   : INC(leftrect.right,controls.list[i](Control).width);
          |  alTop    : INC(toprect.bottom,controls.list[i](Control).height);
          |  alRight  : DEC(rightrect.left,controls.list[i](Control).width);
          |  alBottom : DEC(bottomrect.top,controls.list[i](Control).height);
          ELSE
          END;
        END;
        IF bottomrect.top < toprect.bottom THEN
          bottomrect.bottom := toprect.bottom + bottomrect.bottom - bottomrect.top;
          bottomrect.top := toprect.bottom
        END;
        leftrect.top      := toprect.bottom;
        leftrect.bottom   := bottomrect.top;
        rightrect.top     := toprect.bottom;
        rightrect.bottom  := bottomrect.top;
        IF rightrect.left < leftrect.right THEN
          rightrect.right := leftrect.right + rightrect.right - rightrect.left;
          rightrect.left := leftrect.right;
        END;
        clientrect.left   := leftrect.right;
        clientrect.top    := toprect.bottom;
        clientrect.right  := rightrect.left;
        clientrect.bottom := bottomrect.top;
      END AdjustRects;

    BEGIN
      IF controls # NIL THEN
        W.GetClientRect(handle,rect);

        INC(rect.top, border);
        INC(rect.left, border);
        DEC(rect.right, border);
        DEC(rect.bottom, border);

        AdjustRects;
        lw := 0;
        th := 0;
        rw := 0;
        bh := 0;
        flag := FALSE;
        FOR i := 0 TO controls.count - 1 DO

          CASE controls.list[i](Control).align OF
          |  alLeft   :
                        newleft   := leftrect.left + lw;
                        lw        := lw + controls.list[i](Control).width;
                        newtop    := leftrect.top;
                        newwidth  := controls.list[i](Control).width;
                        newheight := leftrect.bottom - leftrect.top;
          |  alTop    :
                        newleft   := toprect.left;
                        newtop    := th + toprect.top;
                        th        := th + controls.list[i](Control).height;
                        newwidth  := toprect.right - toprect.left;
                        newheight := controls.list[i](Control).height;
          |  alRight  :
                        newleft   := rightrect.right - rw - controls.list[i](Control).width;
                        rw        := rw + controls.list[i](Control).width;
                        newtop    := rightrect.top;
                        newwidth  := controls.list[i](Control).width;
                        newheight := rightrect.bottom - rightrect.top;
          |  alBottom :
                        newleft   := bottomrect.left;
                        newtop    := bottomrect.bottom - bh - controls.list[i](Control).height;
                        bh        := bh + controls.list[i](Control).height;
                        newwidth  := bottomrect.right - bottomrect.left;
                        newheight := controls.list[i](Control).height;
          |  alClient :
                        newleft   := leftrect.right;
                        newtop    := toprect.bottom;
                        newwidth  := rightrect.left - leftrect.right;
                        newheight := bottomrect.top - toprect.bottom;

          ELSE
          END;

          IF controls.list[i](Control).align # alNone THEN
            controls.list[i](Control).left   := newleft;
            controls.list[i](Control).top    := newtop;
            controls.list[i](Control).width  := newwidth;
            controls.list[i](Control).height := newheight;
          END;

          controls.list[i](Control).CheckState;

        END;

      END;

    END AlignControls;

    PROCEDURE CheckState*;
    VAR
      i: Longint;
    BEGIN

      IF handle = 0 THEN
        CreateControl;
      ELSE
        IF controls # NIL THEN
          FOR i := 0 TO controls.count - 1 DO
            controls.list[i](Control).CheckState;
          END;
        END;
        statedata := {};
        IF state.position    # position    THEN INCL(statedata,1) END;
        IF state.left        # left        THEN INCL(statedata,4) END;
        IF state.top         # top         THEN INCL(statedata,4) END;
        IF state.width       # width       THEN INCL(statedata,4) END;
        IF state.height      # height      THEN INCL(statedata,4) END;
        IF state.visible     # visible     THEN INCL(statedata,5) END;
        IF state.enabled     # enabled     THEN INCL(statedata,6) END;
        IF state.color       # color       THEN INCL(statedata,7) END;

        IF statedata # {} THEN

          IF 1 IN statedata THEN
            IF owner = NIL THEN
              SetPosition;
              W.MoveWindow(handle,left,top,width,height,0);
            END;
          END;

          IF 4 IN statedata THEN
            W.MoveWindow(handle,left,top,width,height,1);
          END;

          IF 5 IN statedata THEN
            IF ~ visible THEN
              W.ShowWindow(handle, W.SW_HIDE);
              WMHide;
            ELSE
              W.ShowWindow(handle, W.SW_SHOW);
              WMShow;
            END;
          END;

          IF 6 IN statedata THEN
            IF ~ enabled THEN
              WMDisable;
              W.EnableWindow(handle,W.False);
            ELSE
              WMEnable;
              W.EnableWindow(handle,W.True);
              IF brush # 0 THEN W.DeleteObject(brush) END;
              brush := W.CreateSolidBrush(color);
              W.SetClassLong(handle,W.GCL_HBRBACKGROUND,brush);
              W.InvalidateRect(handle,NIL,W.True);
            END;
          END;

          IF 7 IN statedata THEN
            IF brush # 0 THEN W.DeleteObject(brush) END;
            brush := W.CreateSolidBrush(color);
            W.SetClassLong(handle,W.GCL_HBRBACKGROUND,brush);
            W.InvalidateRect(handle,NIL,W.True);
          END;

          SaveState;

        END;

      END;

    END CheckState;

    PROCEDURE SaveState*;
    BEGIN
      state.position    := position;
      state.left        := left;
      state.top         := top;
      state.width       := width;
      state.height      := height;
      state.visible     := visible;
      state.enabled     := enabled;
      state.color       := color;
      statedata         := {};
    END SaveState;

    PROCEDURE Hide*;
    BEGIN
      visible := FALSE;
    END Hide;

    PROCEDURE Show*;
    BEGIN
      visible := TRUE;
    END Show;

    PROCEDURE Close*;
    VAR
      message: Message;
    BEGIN
      message.msg := W.WM_CLOSE;
      WindowProc(message);
    END Close;

    PROCEDURE BeforeCreate*;
    BEGIN
    END BeforeCreate;

    PROCEDURE AfterCreate*;
    BEGIN
    END AfterCreate;

    PROCEDURE WMEnable*;
    BEGIN
    END WMEnable;

    PROCEDURE WMDisable*;
    BEGIN
    END WMDisable;

    PROCEDURE WMShow*;
    BEGIN
    END WMShow;

    PROCEDURE WMHide*;
    BEGIN
    END WMHide;

    PROCEDURE WMMouseLeave*;
    BEGIN
    END WMMouseLeave;

    PROCEDURE WMMouseEnter*;
    BEGIN
    END WMMouseEnter;

    PROCEDURE WMMouseHover*;
    BEGIN
    END WMMouseHover;

    PROCEDURE WMCommand*(VAR message: Message);
    BEGIN
    END WMCommand;

    PROCEDURE WMMouseMove*(VAR message: Message);
    BEGIN
    END WMMouseMove;

    PROCEDURE WMMouseClick*(VAR message: Message);
    BEGIN
      IF OnClick # NIL THEN OnClick(SELF); END;
    END WMMouseClick;

    PROCEDURE WMMouseDown*(VAR message: Message);
    BEGIN
    END WMMouseDown;

    PROCEDURE WMMouseUp*(VAR message: Message);
    BEGIN
    END WMMouseUp;

    PROCEDURE WMMove*(VAR message: Message);
    BEGIN
    END WMMove;

    PROCEDURE WMSize*(VAR message: Message);
    BEGIN
    END WMSize;

    PROCEDURE DrawCaption*(dc: W.HDC);
    VAR
      size : W.Size;
      old  : Longint;
      rect : W.Rect;
    BEGIN
      IF owner # NIL THEN
        old := W.SetBkMode(dc,W.TRANSPARENT);
        W.SetTextColor(dc,font.color);
        W.SelectObject(dc, font.handle);
        W.GetTextExtentPoint32(dc,SYSTEM.ADR(caption),Cl.Length(caption),size);
        rect.left   := (width - size.cx) DIV 2 - 1;
        rect.top    := (height - size.cy) DIV 2;
        rect.right  := rect.left + size.cx + 2;
        rect.bottom := rect.top + size.cy;
        W.FillRect(dc,rect,brush);
        W.ExtTextOut(dc,rect.left+1,rect.top,W.ETO_CLIPPED,NIL,SYSTEM.ADR(caption),Cl.Length(caption), NIL);
        W.SetBkMode(dc,old);
      END;
    END DrawCaption;

    PROCEDURE WMPaint*(VAR message: Message);
    VAR
      dc    : W.HDC;
    BEGIN
      dc := W.GetDC(handle);
      DrawCaption(dc);
      W.ReleaseDC(handle,dc);
    END WMPaint;

    PROCEDURE WMEraseBkgnd*(VAR message: Message);
    BEGIN
    END WMEraseBkgnd;

    PROCEDURE WindowProc*(VAR message: Message);
    VAR
      p,m    : W.Point;
      x,y    : Longint;
      control: Control;
      tme    : W.TrackMouseEventEx;
    BEGIN

        CASE message.msg OF
        |  W.WM_LBUTTONDOWN : mousedown := TRUE; message.button := mbLeft;
        |  W.WM_LBUTTONUP   : message.button := mbLeft;
        |  W.WM_RBUTTONDOWN : mousedown := TRUE; message.button := mbRight;
        |  W.WM_RBUTTONUP   : message.button := mbRight;
        |  W.WM_MBUTTONDOWN : mousedown := TRUE; message.button := mbMiddle;
        |  W.WM_MBUTTONUP   : message.button := mbMiddle;
        ELSE
        END;

        CASE message.msg OF
        |  W.WM_LBUTTONDOWN : WMMouseDown(message);
        |  W.WM_LBUTTONUP   : WMMouseUp(message);
        |  W.WM_RBUTTONDOWN : WMMouseDown(message);
        |  W.WM_RBUTTONUP   : WMMouseUp(message);
        |  W.WM_MBUTTONDOWN : WMMouseDown(message);
        |  W.WM_MBUTTONUP   : WMMouseUp(message);
        |  W.WM_PAINT       :
                              IF message.dc = 0 THEN
                                W.DefWindowProc(handle, message.msg, message.wparam, message.lparam);
                                message.dc := 1;
                                WindowProc(message);
                              ELSE
                                WMPaint(message);
                                message.dc     := 0;
                                message.result := 0;
                              END;
        |  W.WM_ERASEBKGND  : WMEraseBkgnd(message);
                              IF message.result = 0 THEN
                                W.DefWindowProc(handle, message.msg, message.wparam, message.lparam);
                              END;
        |  W.WM_MOUSEMOVE   :
                              tme.cbSize := SIZE(W.TrackMouseEventEx);
                              tme.hwndTrack := handle;
                              tme.dwFlags := W.TME_LEAVE+W.TME_HOVER;
                              W.TrackMouseEvent(tme);
                              IF ~ mouseinside THEN
                                mouseinside := TRUE;
                                WMMouseEnter;
                              END;
                              WMMouseMove(message);
                              IF showhint THEN
                                x := LoWord(message.lparam);
                                y := HiWord(message.lparam);
                                IF (mainhint # NIL) & mainhint.visible THEN
                                  IF (forhover.x # x) OR (forhover.y # y) THEN
                                    mainhint.visible := FALSE;
                                  END;
                                END;
                                forhover.x := x;
                                forhover.y := y;
                              END;
        |  W.WM_MOUSELEAVE :
                              mouseinside := FALSE;
                              mousehover  := FALSE;
                              WMMouseLeave;
        |  W.WM_MOUSEHOVER :
                              IF ~ mousehover THEN
                                mousehover := TRUE;
                                WMMouseHover;
                                IF showhint THEN
                                  IF (mainhint # NIL) & (hint # "") THEN
                                    (*IF mainhint.visible THEN HideHint END;*)
                                    W.GetCursorPos(p);
                                    mainhint.text := hint;
                                    W.SetWindowPos(mainhint.handle,W.HWND_TOPMOST,p.x + 2,p.y - mainhint.height - 2,mainhint.width,mainhint.height,W.SWP_SHOWWINDOW + W.SWP_NOACTIVATE);
                                    mainhint.visible       := TRUE;
                                    mainhint.state.visible := TRUE;
                                  END;
                                END;
                              END;
        |  W.WM_MOVE	  :
                              WMMove(message);
                              UpdateBounds;
        |  W.WM_SIZE	  :
                              WMSize(message);
                              UpdateBounds;
                              AlignControls;
        |  W.WM_COMMAND	  :
                              IF controls # NIL THEN
                                control := controls.list[message.wparam](Control);
                                IF control # NIL THEN
                                  control.WMCommand(message);
                                END;
                              END;
        |  W.WM_CLOSE       :
                              IF SELF = mainform THEN
                                W.PostQuitMessage(0);
                              ELSE
                                SELF.Hide;
                              END;
        |  W.WM_DESTROY     :
                              IF brush # 0 THEN W.DeleteObject(brush) END;
                              IF font.handle # 0 THEN W.DeleteObject(font.handle) END;
        ELSE
          message.result := W.DefWindowProc(handle, message.msg, message.wparam, message.lparam);
        END;

        CASE message.msg OF
        |  W.WM_LBUTTONDOWN,
           W.WM_RBUTTONDOWN,
           W.WM_MBUTTONDOWN : W.SetCapture(handle);

        |  W.WM_LBUTTONUP,
           W.WM_RBUTTONUP,
           W.WM_MBUTTONUP   :
                              IF mousedown THEN
                                W.GetCursorPos(p);
                                W.ClientToScreen(handle,m);
                                IF (p.x>=m.x) & (p.y>=m.y) & (p.x<m.x+width) & (p.y<m.y+height) THEN
                                  WMMouseClick(message);
                                END;
                              END;
                              W.ReleaseCapture();
                              mousedown := FALSE;
        ELSE
        END;

    END WindowProc;

  END;

  Hint* = POINTER TO RECORD (Control)

    text*   : Name;

    PROCEDURE & Create*(Parent: Control);
    BEGIN
      Create^(Parent);
      classname    := "Hint";
      visible      := FALSE;
      width        := 1;
      height       := 16;
      left         := 0;
      top          := 0;
      color        := G.clWebLightYellow;
    END Create;

    PROCEDURE CreateParams*;
    BEGIN
      CreateParams^;
      style   := W.WS_POPUP;
      exstyle := W.WS_EX_TOOLWINDOW;
      wndstyle := wndstyle - W.CS_VREDRAW - W.CS_HREDRAW + W.CS_DROPSHADOW;
    END CreateParams;

    PROCEDURE WMPaint*(VAR message: Message);
    VAR
      dc   : W.HDC;
      rect : W.Rect;
      size : W.Size;
      font : W.HFont;
      adr  : Longint;
      len  : Longint;
    BEGIN
      IF text # "" THEN
        dc := W.GetDC(handle);
        font := W.GetStockObject(W.DEFAULT_GUI_FONT);
        W.SelectObject(dc,font);
        W.SetBkMode(dc,W.TRANSPARENT);
        adr := SYSTEM.ADR(text);
        len := Cl.Length(text);
        W.GetTextExtentPoint32(dc,adr,len,size);

        IF (width # size.cx + 8) OR (height # size.cy + 4) THEN
          width  := size.cx + 8;
          height := size.cy + 4;
          W.SetWindowPos(handle,W.HWND_TOPMOST,0,0,width,height,W.SWP_NOMOVE + W.SWP_NOACTIVATE);
          state.width  := width;
          state.height := height;
        END;

        G.Rectangle(dc,0,0,width-1,height-1);

        W.SetTextColor(dc,G.clBlack);
        W.ExtTextOut(dc,4,(height - size.cy) DIV 2,W.ETO_CLIPPED,NIL,adr,len,NIL);

        W.DeleteObject(font);
        W.ReleaseDC(handle,dc);
      END;
    END WMPaint;

  END;

  Customcontrol* = POINTER TO RECORD (Control)

    canvas*: G.Bitmap;

    PROCEDURE & Create*(Parent: Control);
    BEGIN
      Create^(Parent);
      classname    := "Customcontrol";
      width        := 250;
      height       := 250;
      NEW(canvas);
      canvas.SetSize(width,height);
    END Create;

    PROCEDURE Destroy*;
    BEGIN
      canvas.Free;
      Destroy^;
    END Destroy;

    PROCEDURE CreateParams*;
    BEGIN
      CreateParams^;
      style     := style - W.WS_BORDER;
    END CreateParams;

    PROCEDURE WMSize*(VAR message: Message);
    BEGIN
      WMSize^(message);
      canvas.SetSize(width,height);
    END WMSize;

    PROCEDURE WMEraseBkgnd*(VAR message: Message);
    BEGIN
      message.result := 1;
    END WMEraseBkgnd;

    PROCEDURE Paint*;
    BEGIN
    END Paint;

    PROCEDURE WMPaint*(VAR message: Message);
    VAR
      dc : W.HDC;
    BEGIN
      dc := W.GetDC(handle);
      canvas.FillRect(0,0,width,height,color);
      Paint;
      canvas.Draw(dc,0,0);
      W.ReleaseDC(handle,dc);
    END WMPaint;

  END;

PROCEDURE [WINAPI] WndProc(hwnd: W.HWND; uMsg: Longint; wParam: W.WParam; lParam: W.LParam): W.LResult;
VAR
  control: Control;
  message: Message;
BEGIN

  message.msg    := uMsg;
  message.wparam := wParam;
  message.lparam := lParam;
  message.result := 0;
  message.button := 0;
  message.dc     := 0;

  control := SYSTEM.VAL(Control,W.GetProp(hwnd,SYSTEM.ADR("Control")));
  IF control # NIL THEN
    control.WindowProc(message);
  ELSE
    RETURN W.DefWindowProc(hwnd, uMsg, wParam, lParam);
  END;
  RETURN message.result;

END WndProc;

PROCEDURE LoWord*(VAR n: Longint): Longint;
BEGIN
  RETURN ASH(SYSTEM.LSH(n, 16), -16)
END LoWord;

PROCEDURE HiWord*(VAR n: Longint): Longint;
BEGIN
  RETURN ASH(n, -16);
END HiWord;

PROCEDURE XYInRect*(x,y: Longint; rect: W.Rect): Boolean;
BEGIN
  RETURN (x >= rect.left) & (x < rect.right) & (y >= rect.top) & (y < rect.bottom);
END XYInRect;

PROCEDURE Run*;
VAR
  msg    : W.Msg;
  i      : Longint;
BEGIN

  FOR i:=0 TO screen.count - 1 DO
    screen.list[i](Control).SetPosition;
    screen.list[i](Control).CreateControl;
  END;

  WHILE W.GetMessage(msg,0,0,0) # 0 DO
    W.TranslateMessage(msg);
    W.DispatchMessage(msg);
    FOR i:=0 TO screen.count - 1 DO
      IF screen.list[i](Control).enabled THEN
        screen.list[i](Control).CheckState;
      END;
    END;
  END;

  FOR i:=0 TO screen.count - 1 DO
    screen.list[i](Control).Destroy;
  END;
  screen.Clear;
  screen := NIL;

  Cl.ClassList.UnregisterClass;
  W.ExitProcess(0);
END Run;

END Controls.
