MODULE Dialogs;

IMPORT
  SYSTEM, Kernel, W := Windows, C := Controls;

TYPE
  Char         = W.Char;
  Boolean      = W.Boolean;
  Shortint     = W.Shortint;
  Integer      = W.Integer;
  Longint      = W.Longint;
  Real         = W.Real;
  Longreal     = W.Longreal;
  Pointer      = W.Pointer;
  Set	       = W.Set;
  String       = W.String;
  Name         = W.Name;

  Opendialog* = POINTER TO RECORD
    FileName*    : ARRAY 256 OF Char;
    Filter*      : ARRAY 256 OF Char;
    Ext*         : ARRAY 10  OF Char;
    FilterIndex* : Longint;

    dialog       : W.OpenFileNameA;

    PROCEDURE & Create*;
    BEGIN
      FileName[0] := 0X;
      FilterIndex := 1;
    END Create;

    PROCEDURE Convert (VAR a: String);
    VAR
      i: Longint;
    BEGIN
      i := 0;
      WHILE a[i] # 0X DO
        IF a[i]='|' THEN a[i]:=0X END;
        INC(i);
      END;
    END Convert;

    PROCEDURE FillRec();
    BEGIN
      dialog.lStructSize       := SIZE(W.OpenFileNameA);
      dialog.hwndOwner         := 0;
      dialog.hInstance         := W.hInstance;
      dialog.lpstrFilter       := SYSTEM.ADR(Filter);
      dialog.lpstrCustomFilter := 0;
      dialog.nMaxCustFilter    := 0;
      dialog.nFilterIndex      := FilterIndex;
      dialog.lpstrFile         := SYSTEM.ADR(FileName);
      dialog.nMaxFile          := LEN(FileName);
      dialog.lpstrFileTitle    := W.NULL;
      dialog.nMaxFileTitle     := 0;
      dialog.lpstrInitialDir   := W.NULL;
      dialog.lpstrTitle        := W.NULL;
      dialog.Flags             := W.OFN_ALLOWMULTISELECT + W.OFN_EXPLORER + W.OFN_PATHMUSTEXIST + W.OFN_FILEMUSTEXIST;
      dialog.nFileOffset       := 0;
      dialog.nFileExtension    := 0;
      dialog.lpstrDefExt       := SYSTEM.ADR(Ext);
      dialog.lCustData         := 0;
      dialog.lpfnHook          := NIL;
      dialog.lpTemplateName    := W.NULL;

    END FillRec;

    PROCEDURE Execute* (): Boolean;
    BEGIN
      Convert(Filter);
      FillRec();
      IF W.GetOpenFileName(dialog) # 0 THEN
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END Execute;

  END;

  Savedialog* = POINTER TO RECORD(Opendialog)

    PROCEDURE Execute*(): Boolean;
    BEGIN
      Convert(Filter);
      FillRec^();
      IF W.GetSaveFileName(dialog) # 0 THEN
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END Execute;

  END;

  Colordialog* = POINTER TO RECORD

    color*    : Longint;
    dialog    : W.ChooseColorA;
    acrCustClr: ARRAY 16 OF W.ColorRef;

    PROCEDURE & Create*;
    BEGIN
      dialog.lStructSize     := SIZE(W.ChooseColorA);
      dialog.hwndOwner       := 0;
      dialog.hInstance       := 0;
      dialog.rgbResult       := color;
      dialog.lpCustColors    := SYSTEM.ADR(acrCustClr);
      dialog.Flags           := W.CC_FULLOPEN + W.CC_RGBINIT;
      dialog.lCustData       := 0;
      dialog.lpfnHook        := NIL;
      dialog.lpTemplateName  := W.NULL;
    END Create;

    PROCEDURE Execute* (): BOOLEAN;
    BEGIN
      IF W.ChooseColor(dialog) # 0 THEN
        color := dialog.rgbResult;
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END Execute;

  END;

  Fontdialog* = POINTER TO RECORD
    font*      : W.HFont;
    fontname*  : ARRAY 32 OF CHAR;
    color*     : LONGINT;

    dialog     : W.ChooseFontA;
    logfont    : W.LogFontA;

    PROCEDURE & Create*;
    BEGIN
      dialog.lStructSize := SIZE(W.ChooseFontA);
      dialog.hwndOwner   := 0;
      dialog.lpLogFont   := SYSTEM.ADR(logfont);
      dialog.rgbColors   := color;
      dialog.Flags       := W.CF_SCREENFONTS + W.CF_EFFECTS;
    END Create;

    PROCEDURE Execute* (): BOOLEAN;
    BEGIN
      IF W.ChooseFont(dialog )# 0 THEN
        font := W.CreateFontIndirect(logfont);
        SYSTEM.MOVE(SYSTEM.ADR(logfont.lfFaceName),SYSTEM.ADR(fontname),LEN(logfont.lfFaceName));
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END Execute;

  END;

  Printdialog* = POINTER TO RECORD
    dialog: W.PrintDlgA;

    PROCEDURE & Create*;
    BEGIN
      dialog.lStructSize        := SIZE(W.PrintDlgA)-2;
      dialog.hwndOwner          := 0;
      dialog.hDevMode           := 0;
      dialog.hDevNames          := 0;
      dialog.hDC                := 0;
      dialog.Flags              := W.PD_USEDEVMODECOPIESANDCOLLATE + W.PD_RETURNDC;
      dialog.nFromPage          := -1;
      dialog.nToPage            := -1;
      dialog.nMinPage           :=  1;
      dialog.nMaxPage           := -1;
      dialog.nCopies            :=  1;
      dialog.hInstance          :=  0;
      dialog.lCustData          := 0;
      dialog.lpfnPrintHook      := NIL;
      dialog.lpfnSetupHook      := NIL;
      dialog.lpPrintTemplateName:= W.NULL;
      dialog.lpSetupTemplateName:= W.NULL;
      dialog.hPrintTemplate     := 0;
      dialog.hSetupTemplate     := 0;

    END Create;

    PROCEDURE Execute* (): BOOLEAN;
    BEGIN
      IF W.PrintDlg(dialog) # 0 THEN
        W.DeleteDC(dialog.hDC);
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END Execute;

  END;

END Dialogs.

